diff --git a/src/openvpn/options.c b/src/openvpn/options.c
index cf971a6..eaca84b 100644
--- a/src/openvpn/options.c
+++ b/src/openvpn/options.c
@@ -647,6 +647,7 @@ static const char usage_message[] =
   "                                  cache until token is removed.\n"
   "--pkcs11-id-management          : Acquire identity from management interface.\n"
   "--pkcs11-id serialized-id 'id'  : Identity to use, get using standalone --show-pkcs11-ids\n"
+  "--pkcs11-pinfile 'file'         : Read PKCS11 pin from file\n"
 #endif			/* ENABLE_PKCS11 */
  "\n"
   "SSL Library information:\n"
@@ -1711,6 +1712,7 @@ show_settings (const struct options *o)
   SHOW_INT (pkcs11_pin_cache_period);
   SHOW_STR (pkcs11_id);
   SHOW_BOOL (pkcs11_id_management);
+  SHOW_STR (pkcs11_pinfile);
 #endif			/* ENABLE_PKCS11 */
 
 #if P2MP
@@ -2401,6 +2403,7 @@ options_postprocess_verify_ce (const struct options *options, const struct conne
       MUST_BE_UNDEF (pkcs11_private_mode[0]);
       MUST_BE_UNDEF (pkcs11_id);
       MUST_BE_UNDEF (pkcs11_id_management);
+      MUST_BE_UNDEF (pkcs11_pinfile);
 #endif
 
       if (pull)
@@ -7276,6 +7279,11 @@ add_option (struct options *options,
       VERIFY_PERMISSION (OPT_P_GENERAL);
       options->pkcs11_id_management = true;
     }
+  else if (streq (p[0], "pkcs11-authfile") && p[1] && !p[2])
+    {
+      VERIFY_PERMISSION (OPT_P_GENERAL);
+      options->pkcs11_pinfile = p[1];
+    }
 #endif
   else if (streq (p[0], "rmtun") && !p[1])
     {
diff --git a/src/openvpn/options.h b/src/openvpn/options.h
index 7bb36c9..cd066d1 100644
--- a/src/openvpn/options.h
+++ b/src/openvpn/options.h
@@ -527,6 +527,7 @@ struct options
   int pkcs11_pin_cache_period;
   const char *pkcs11_id;
   bool pkcs11_id_management;
+  const char *pkcs11_pinfile;
 #endif
 
 #ifdef ENABLE_CRYPTOAPI
diff --git a/src/openvpn/pkcs11.c b/src/openvpn/pkcs11.c
index a1f13c5..efdd6b2 100644
--- a/src/openvpn/pkcs11.c
+++ b/src/openvpn/pkcs11.c
@@ -221,7 +221,6 @@ _pkcs11_openvpn_pin_prompt (
 	char prompt[1024];
 
 	(void)global_data;
-	(void)user_data;
 	(void)retry;
 
 	ASSERT (token!=NULL);
@@ -234,7 +233,7 @@ _pkcs11_openvpn_pin_prompt (
 	if (
 		!get_user_pass (
 			&token_pass,
-			NULL,
+			user_data, /* pinfile as user_data pointer */
 			prompt,
 			GET_USER_PASS_MANAGEMENT|GET_USER_PASS_PASSWORD_ONLY|GET_USER_PASS_NOFATAL
 		)
@@ -610,7 +609,8 @@ int
 tls_ctx_use_pkcs11 (
 	struct tls_root_ctx * const ssl_ctx,
 	bool pkcs11_id_management,
-	const char * const pkcs11_id
+	const char * const pkcs11_id,
+	const char * const pkcs11_pinfile
 ) {
 	pkcs11h_certificate_id_t certificate_id = NULL;
 	pkcs11h_certificate_t certificate = NULL;
@@ -678,7 +678,7 @@ tls_ctx_use_pkcs11 (
 	if (
 		(rv = pkcs11h_certificate_create (
 			certificate_id,
-			NULL,
+			pkcs11_pinfile, /* pinfile as user_data pointer */
 			PKCS11H_PROMPT_MASK_ALLOW_ALL,
 			PKCS11H_PIN_CACHE_INFINITE,
 			&certificate
diff --git a/src/openvpn/pkcs11.h b/src/openvpn/pkcs11.h
index b49401c..66e0d4c 100644
--- a/src/openvpn/pkcs11.h
+++ b/src/openvpn/pkcs11.h
@@ -63,7 +63,8 @@ int
 tls_ctx_use_pkcs11 (
 	struct tls_root_ctx * const ssl_ctx,
 	bool pkcs11_id_management,
-	const char * const pkcs11_id
+	const char * const pkcs11_id,
+	const char * const pkcs11_pinfile
 );
 
 void
diff --git a/src/openvpn/ssl.c b/src/openvpn/ssl.c
index a48c8ec..a8efac0 100644
--- a/src/openvpn/ssl.c
+++ b/src/openvpn/ssl.c
@@ -525,7 +525,7 @@ init_ssl (const struct options *options, struct tls_root_ctx *new_ctx)
 #ifdef ENABLE_PKCS11
   else if (options->pkcs11_providers[0])
     {
-      if (!tls_ctx_use_pkcs11 (new_ctx, options->pkcs11_id_management, options->pkcs11_id))
+      if (!tls_ctx_use_pkcs11 (new_ctx, options->pkcs11_id_management, options->pkcs11_id, options->pkcs11_pinfile))
 	{
 	  msg (M_WARN, "Cannot load certificate \"%s\" using PKCS#11 interface",
 	      options->pkcs11_id);
